'''Class with methods to validated the current status of the IviumSoft environment'''
from .core import Core
from .errors import DriverNotOpenError, \
    IviumSoftNotRunningError, \
    DeviceNotConnectedToIviumSoftError, \
    DeviceBusyError, \
    NoDeviceDetectedError, \
    CellOffError


class PyviumVerifiers:
    '''Encapsulates methods to validated the current status of the IviumSoft environment'''

    @staticmethod
    def verify_driver_is_open():
        '''Raise exception if the driver is not open'''
        if not Core.is_driver_open():
            raise DriverNotOpenError

    @staticmethod
    def verify_iviumsoft_is_running():
        '''Raise exception if IviumSoft is not running'''
        device_status = Core.IV_getdevicestatus()

        if device_status == -1:
            raise IviumSoftNotRunningError

    @staticmethod
    def verify_device_is_connected_to_iviumsoft():
        '''Raise exception if a device is not connected through the IviumSoft app'''
        device_status = Core.IV_getdevicestatus()
        if device_status < 1 or device_status == 3:
            raise DeviceNotConnectedToIviumSoftError

    @staticmethod
    def veryfy_device_is_connected_to_computer():
        '''Raise exception if no device is connected to your computer through usb'''
        device_status = Core.IV_getdevicestatus()

        if device_status == 3:
            raise NoDeviceDetectedError

    @staticmethod
    def verify_device_is_available():
        '''Raise exception if the connected device/s are not available at the moment'''
        device_status = Core.IV_getdevicestatus()
        if device_status == 2:
            raise DeviceBusyError

    @staticmethod
    def verify_cell_is_on():
        '''Raise exception if the is off'''
        _, device_status = Core.IV_getcellstatus()
        if not device_status:
            raise CellOffError



from .constants import DOUBLE_PTR, LONG_PTR
from .core_base import CoreBase, ffi

ffi.cdef("""
    long __stdcall IV_we32setchannel(long *index);
    long __stdcall IV_we32setoffset(long *index, double *value);
    long __stdcall IV_we32setoffsets(long *nval, double *values);
    long __stdcall IV_we32getoffsets(long *nval, double *values);
    long __stdcall IV_we32readcurrents(double *values);
""")


class We32Functions(CoreBase):
    @staticmethod
    def IV_we32setchannel(channel_index: int) -> int:
        '''Select active WE32 channel (chan)'''
        channel_index_ptr = ffi.new(LONG_PTR, channel_index)
        result_code = CoreBase.get_lib().IV_we32setchannel(channel_index_ptr)
        return result_code

    @staticmethod
    def IV_we32setoffset(channel_index: int, value: float) -> int:
        '''Set WE32 offset (chan,value), value -2 to +2V.
            Use chan=0 to apply the same offset to all channels.'''
        channel_index_ptr = ffi.new(LONG_PTR, channel_index)
        value_ptr = ffi.new(DOUBLE_PTR, value)
        result_code = CoreBase.get_lib().IV_we32setoffset(channel_index_ptr, value_ptr)
        return result_code

    @staticmethod
    def IV_we32setoffsets(number_of_channels: int, value: float) -> int:
        '''REVISE! Set WE32 offsets values (Nchan,values),
            with Nchan the number of channels (1..32)'''
        number_of_channels_index_ptr = ffi.new(LONG_PTR, number_of_channels)
        value_ptr = ffi.new(DOUBLE_PTR, value)
        result_code = CoreBase.get_lib().IV_we32setoffsets(
            number_of_channels_index_ptr, value_ptr)
        return result_code

    @staticmethod
    def IV_we32getoffsets(number_of_channels: int) -> tuple[int, float]:
        '''REVISE! Returns actual WE32 offset values (Nchan,values),
            with Nchan the number of channels (1..32)'''
        number_of_channels_index_ptr = ffi.new(LONG_PTR, number_of_channels)
        values_ptr = ffi.new(DOUBLE_PTR)
        result_code = CoreBase.get_lib().IV_we32getoffsets(
            number_of_channels_index_ptr, values_ptr)
        return result_code, values_ptr[0]

    @staticmethod
    def IV_we32readcurrents() -> tuple[int, float]:
        '''REVISE! Returns array with 32 WE32 current values,
            that are measured simultaneously'''
        current_values_ptr = ffi.new(DOUBLE_PTR)
        result_code = CoreBase.get_lib().IV_we32readcurrents(current_values_ptr)
        return result_code, current_values_ptr[0]



import os
def get_file_list(dir_path='.'):
    '''Obtains a list of files in a directory'''
    file_list= []
    normalized_path = os.path.normpath(dir_path)
    for path in os.listdir(normalized_path):
        if os.path.isfile(os.path.join(normalized_path, path)):
            file_list.append(path)
    
    return file_list


from ..core import Core
from ..errors import (DeviceNotConnectedToIviumSoftError,
                      IviumSoftNotRunningError)
from ..pyvium_verifiers import PyviumVerifiers


class GenericFunctions():
    @staticmethod
    def open_driver():
        '''Open the driver to manipulate the Ivium software'''
        if Core.is_driver_open():
            Core.IV_close()
        Core.IV_open()
        try:
            PyviumVerifiers.verify_iviumsoft_is_running()
        except:
            Core.IV_close()
            raise

    @staticmethod
    def close_driver():
        '''Closes the iviumSoft driver'''
        PyviumVerifiers.verify_driver_is_open()
        Core.IV_close()

    @staticmethod
    def get_max_device_number():
        '''Returns the maximum number of devices that can be managed by IviumSoft'''
        PyviumVerifiers.verify_driver_is_open()
        return Core.IV_MaxDevices()

    @staticmethod
    def get_device_status() -> tuple[int, str]:
        '''It returns -1 (no IviumSoft), 0 (not connected), 1 (available_idle), 2 (available_busy),
            3 (no device available)'''
        PyviumVerifiers.verify_driver_is_open()
        PyviumVerifiers.verify_iviumsoft_is_running()
        status_labels = {
            '-1': 'no IviumSoft',
            '0': 'connected',
            '1': 'available_idle',
            '2': 'available_busy',
            '3': 'no device available'
        }
        result_code = Core.IV_getdevicestatus()
        return result_code, status_labels[str(result_code)]

    @staticmethod
    def is_iviumsoft_running() -> bool:
        '''It returns true if if the selected instance of IviumSoft is running'''
        PyviumVerifiers.verify_driver_is_open()
        return Core.IV_getdevicestatus() != -1

    @staticmethod
    def get_active_iviumsoft_instances():
        '''Returns a list of active(open) IviumSoft instances'''
        PyviumVerifiers.verify_driver_is_open()
        active_instances = []
        first_active_instance_number = 0
        for instance_number in range(1, 32):
            Core.IV_selectdevice(instance_number)

            if Core.IV_getdevicestatus() != -1:
                active_instances.append(instance_number)

                if first_active_instance_number == 0:
                    first_active_instance_number = instance_number

        if first_active_instance_number == 0:
            first_active_instance_number = 1

        Core.IV_selectdevice(first_active_instance_number)
        return active_instances

    @staticmethod
    def select_iviumsoft_instance(iviumsoft_instance_number: int):
        '''It allows to select one instance of the currently running IviumSoft instances'''

        PyviumVerifiers.verify_driver_is_open()
        active_instances = GenericFunctions.get_active_iviumsoft_instances()
        if iviumsoft_instance_number not in active_instances:
            error_msg = 'No IviumSoft on instance number {}, actual active instances list = {}'
            raise IviumSoftNotRunningError(error_msg.format(
                iviumsoft_instance_number, active_instances))
        Core.IV_selectdevice(iviumsoft_instance_number)

    @staticmethod
    def get_device_serial_number():
        '''Returns the serial number of the currently selected device if available'''
        PyviumVerifiers.verify_driver_is_open()
        PyviumVerifiers.verify_iviumsoft_is_running()
        PyviumVerifiers.veryfy_device_is_connected_to_computer()
        _, serial_number = Core.IV_readSN()
        if serial_number == '':
            raise DeviceNotConnectedToIviumSoftError(
                'This device needs to be connected to get its serial number')
        return serial_number

    @staticmethod
    def connect_device():
        '''It connects the currently selected device'''
        PyviumVerifiers.verify_driver_is_open()
        PyviumVerifiers.verify_iviumsoft_is_running()
        PyviumVerifiers.veryfy_device_is_connected_to_computer()
        Core.IV_connect(1)

    @staticmethod
    def disconnect_device():
        '''It disconnects the currently selected device'''
        PyviumVerifiers.verify_driver_is_open()
        PyviumVerifiers.verify_iviumsoft_is_running()
        PyviumVerifiers.veryfy_device_is_connected_to_computer()
        Core.IV_connect(0)

    @staticmethod
    def get_dll_version() -> int:
        '''Returns the version of the IviumSoft dll'''
        PyviumVerifiers.verify_driver_is_open()
        return Core.IV_VersionDll()

    @staticmethod
    def get_iviumsoft_version() -> str:
        '''Returns the version of the IviumSoft that match with this pyvium version'''
        PyviumVerifiers.verify_driver_is_open()
        version_str = str(Core.IV_VersionDllFile())[slice(5)]
        sliced_str = slice(5)
        version = version_str[sliced_str]
        return version[:1] + '.' + version[1:]

    @staticmethod
    def select_channel(channel_number: int):
        '''Sending the integer value communicates with Multichannel control:
            if not yet active,
            the [int] number of tabs is automatically opened and the [int] tab becomes active;
            if Ivium-n-Soft is active already, the [int] tab becomes active.
            Now the channel/instrument that is connected to this tab can be controlled.
            If no instrument is connected,
            the next available instrument in the list can be connected (IV_connect) and
            controlled.'''
        PyviumVerifiers.verify_driver_is_open()
        PyviumVerifiers.verify_iviumsoft_is_running()
        Core.IV_SelectChannel(channel_number)



import os
import csv
from typing import Dict, List, Union
from tqdm import tqdm

from ..util import get_file_list

class DataProcessing():
    @staticmethod
    def export_to_csv(data,file_path) -> None:
        '''Saves data on a .csv file.'''
        path = os.path.normpath(file_path)
        with open(path, 'w', encoding='UTF-8') as f:
            write = csv.writer(f)
            write.writerows(data)

    @staticmethod
    def get_idf_data(idf_path) -> List[List[int]]:
        '''Extracts the data from a ivium .idf file and returns a lits of points (data matrix)'''
        data = []

        # open and read idf file
        with open(idf_path, 'r', encoding='ISO-8859-2') as idf:
            raw_data = idf.read()
    
        # split the file into a list of lines
        lines = raw_data.splitlines()
    
        for index,line in enumerate(lines):
            if 'primary_data' in line:
                datapoints = int(lines[index+2])
                start = index+3
                end = start + datapoints
                for line_index in range(start, end):
                    row = lines[line_index].split()
                    datapoint = [eval(value) for value in row]            
                    data.append(datapoint)
        return data

    @staticmethod
    def convert_idf_to_csv(idf_path) -> None:
        '''Extracts the data from a ivium .idf file and saves the data to a .csv file'''
        path = os.path.normpath(rf'{idf_path}')
        data = DataProcessing.get_idf_data(path)
        DataProcessing.export_to_csv(data,path+'.csv')
    
    @staticmethod
    def convert_idf_dir_to_csv(idf_dir_path='.') -> Dict[str,Union[int,List[str]]]:
        '''Extracts the data of all .idf files on a directory and saves the data .csv files'''
        path = os.path.normpath(idf_dir_path)
        files = get_file_list(path)
        idf_files = list(filter(lambda file: (file[-4:] == '.idf'), files))
        converted_count = 0
        errors = []
        for idf_filename in tqdm(idf_files):
            try:
                DataProcessing.convert_idf_to_csv(path+'\\'+idf_filename)
                converted_count += 1
            except Exception:
                errors.append(idf_filename)

        return {"converted_count":converted_count, "error_count": len(errors),"errors":errors}




from .constants import CHAR_ARRAY, LONG_PTR, UTF_ENCODING
from .core_base import CoreBase, ffi

ffi.cdef(
    """
    long __stdcall IV_open();
    long __stdcall IV_close();
    long __stdcall IV_MaxDevices();
    long __stdcall IV_selectdevice(long *devnr);
    long __stdcall IV_getdevicestatus();
    long __stdcall IV_readSN(char *sntext);
    long __stdcall IV_connect(long* devconnect);
    long __stdcall IV_VersionHost(long *version);
    long __stdcall IV_VersionDll();
    long __stdcall IV_VersionCheck();
    long __stdcall IV_HostHandle();
    long __stdcall IV_VersionDllFile();
    long __stdcall IV_VersionDllFileStr();
    long __stdcall IV_SelectChannel(long *channel);
    long __stdcall IV_SelectSn(char *sntext);
    long __stdcall IV_getDbFileName(char *fname);
"""
)


class GenericFunctions(CoreBase):
    @staticmethod
    def IV_open() -> int:
        """Open the driver to manipulate the Ivium software"""
        CoreBase.set_driver_open(True)
        return CoreBase.get_lib().IV_open()

    @staticmethod
    def IV_close() -> int:
        """Closes the iviumSoft driver"""
        CoreBase.set_driver_open(False)
        return CoreBase.get_lib().IV_close()

    @staticmethod
    def IV_MaxDevices() -> int:
        """Returns the maximum number of devices that can be managed by IviumSoft"""
        return CoreBase.get_lib().IV_MaxDevices()

    @staticmethod
    def IV_selectdevice(iviumsoft_instance_number: int = 1) -> tuple[int, int]:
        """It allows to select one instance of the currently running IviumSoft instances"""
        instance_number_ptr = ffi.new(LONG_PTR, iviumsoft_instance_number)
        result_code = CoreBase.get_lib().IV_selectdevice(instance_number_ptr)
        return result_code, instance_number_ptr[0]

    @staticmethod
    def IV_getdevicestatus() -> int:
        """It returns -1 (no IviumSoft), 0 (not connected), 1 (available_idle), 2 (available_busy),
        3 (no device available)"""
        return CoreBase.get_lib().IV_getdevicestatus()

    @staticmethod
    def IV_readSN() -> tuple[int, str]:
        """Returns the serial number of the currently selected device"""
        device_serial_number_ptr = ffi.new(CHAR_ARRAY, 16)
        result_code = CoreBase.get_lib().IV_readSN(device_serial_number_ptr)
        return result_code, ffi.string(device_serial_number_ptr).decode(UTF_ENCODING)

    @staticmethod
    def IV_connect(connection_status: int) -> tuple[int, int]:
        """It connects the currently selected device"""
        connection_status_ptr = ffi.new(LONG_PTR, connection_status)
        result_code = CoreBase.get_lib().IV_connect(connection_status_ptr)
        return result_code, connection_status_ptr[0]

    @staticmethod
    def IV_VersionHost(version_host: int) -> tuple[int, int]:
        """REVISE!!! Returns the version Host"""
        version_host_ptr = ffi.new(LONG_PTR, version_host)
        result_code = CoreBase.get_lib().IV_VersionHost(version_host_ptr)
        return result_code, version_host_ptr[0]

    @staticmethod
    def IV_VersionDll() -> int:
        """Returns the version of the IviumSoft dll"""
        return CoreBase.get_lib().IV_VersionDll()

    @staticmethod
    def IV_VersionCheck() -> int:
        """It returns 1 if the selected instance of IviumSoft is running"""
        return CoreBase.get_lib().IV_VersionCheck()

    @staticmethod
    def IV_HostHandle() -> int:
        """REVISE!!! Returns Host Handle"""
        return CoreBase.get_lib().IV_HostHandle()

    @staticmethod
    def IV_VersionDllFile() -> int:
        """REVISE!!! Returns DLL file version"""
        return CoreBase.get_lib().IV_VersionDllFile()

    @staticmethod
    def IV_VersionDllFileStr() -> str:
        """REVISE!!! Returns DLL file version str"""
        return CoreBase.get_lib().IV_VersionDllFileStr()

    @staticmethod
    def IV_SelectChannel(channel_number: int) -> int:
        """Selects a channel:
        - if not yet active, the [int] number of tabs is automatically opened and
          the [int] tab becomes active.
        - if Ivium-n-Soft is active already, the [int] tab becomes active.
        Now the channel/instrument that is connected to this tab can be controlled.
        If no instrument is connected, the next available instrument in the list can
        be connected (IV_connect) and controlled."""
        channel_number_ptr = ffi.new(LONG_PTR, channel_number)
        result_code = CoreBase.get_lib().IV_SelectChannel(channel_number_ptr)
        return result_code

    @staticmethod
    def IV_SelectSn(serial_number: str) -> tuple[int, str]:
        """Selects device by serial number from the list of available devices to connect"""
        serial_number_ptr = ffi.new(CHAR_ARRAY, serial_number.encode(UTF_ENCODING))
        result_code = CoreBase.get_lib().IV_SelectSn(serial_number_ptr)
        return result_code, ffi.string(serial_number_ptr).decode(UTF_ENCODING)

    @staticmethod
    def IV_getDbFileName():
        """Selects device by serial number from the list of available devices to connect"""
        db_path_ptr = ffi.new(CHAR_ARRAY, 16)
        result_code = CoreBase.get_lib().IV_getDbFileName(db_path_ptr)
        return result_code, ffi.string(db_path_ptr).decode(UTF_ENCODING)



from ..core import Core
from ..pyvium_verifiers import PyviumVerifiers


class MethodModeFunctions():
    @staticmethod
    def load_method(method_file_path: str):
        '''Loads method procedure previously saved to a file.
            method_file_path represents the full path to the file.'''
        PyviumVerifiers.verify_driver_is_open()
        PyviumVerifiers.verify_iviumsoft_is_running()

        result_code, _ = Core.IV_readmethod(method_file_path)

        if result_code == 1:
            raise FileNotFoundError

    @staticmethod
    def save_method(method_file_path: str):
        '''Saves currently loaded method procedure to a file.
            method_file_path represents the full path to the new file.'''
        PyviumVerifiers.verify_driver_is_open()
        PyviumVerifiers.verify_iviumsoft_is_running()

        Core.IV_savemethod(method_file_path)

    @staticmethod
    def start_method(method_file_path=''):
        '''Starts a method procedure.
            If method_file_path is an empty string then the presently loaded procedure is started.
            If the full path to a previously saved method is provided
            then the procedure is loaded from the file and started.'''
        PyviumVerifiers.verify_driver_is_open()
        PyviumVerifiers.verify_iviumsoft_is_running()
        PyviumVerifiers.verify_device_is_connected_to_iviumsoft()
        PyviumVerifiers.verify_device_is_available()

        result_code, _ = Core.IV_startmethod(method_file_path)

        if result_code == 1:
            raise FileNotFoundError

    @staticmethod
    def abort_method():
        '''Aborts the ongoing method procedure'''
        PyviumVerifiers.verify_driver_is_open()
        PyviumVerifiers.verify_iviumsoft_is_running()
        PyviumVerifiers.verify_device_is_connected_to_iviumsoft()

        Core.IV_abort()

    @staticmethod
    def save_data(data_file_path: str):
        '''Saves the results of the last method execution into a file.
            data_file_path represents the full path to the new file.
           IMPORTANT: If the path provided is not valid,
           it will close the selected iviumsoft instance.
        '''
        PyviumVerifiers.verify_driver_is_open()
        PyviumVerifiers.verify_iviumsoft_is_running()

        Core.IV_savedata(data_file_path)

    @staticmethod
    def set_method_parameter(parameter_name: str, parameter_value: str):
        '''Allows updating the parameter values for the currently loaded method procedrue.
            It only works for text based parameters and dropdowns (multiple option selectors).'''
        PyviumVerifiers.verify_driver_is_open()
        PyviumVerifiers.verify_iviumsoft_is_running()

        Core.IV_setmethodparameter(
            parameter_name, parameter_value)

    @staticmethod
    def get_available_data_points_number():
        '''Returns actual available number of datapoints: indicates the progress during a run'''
        PyviumVerifiers.verify_driver_is_open()
        PyviumVerifiers.verify_iviumsoft_is_running()

        _, available_data_points_number = Core.IV_Ndatapoints()

        return available_data_points_number

    @staticmethod
    def get_data_point(data_point_index: int):
        '''Get the data from a datapoint with index int, returns 3 values that depend on
            the used technique. For example LSV/CV methods return (E/I/0) Transient methods
            return (time/I,E/0), Impedance methods return (Z1,Z2,freq) etc.'''
        PyviumVerifiers.verify_driver_is_open()
        PyviumVerifiers.verify_iviumsoft_is_running()

        _, value1, value2, value3 = Core.IV_getdata(
            data_point_index)

        return value1, value2, value3

    @staticmethod
    def get_data_point_from_scan(data_point_index: int, scan_index: int):
        '''Same as get_data_point, but with the additional scan_index parameter.
            This function will allow reading data from non-selected (previous) scans.'''
        PyviumVerifiers.verify_driver_is_open()
        PyviumVerifiers.verify_iviumsoft_is_running()

        _, value1, value2, value3 = Core.IV_getdatafromline(
            data_point_index, scan_index)

        return value1, value2, value3



from .direct_mode_functions import DirectModeFunctions
from .generic_functions import GenericFunctions
from .method_mode_functions import MethodModeFunctions


class Pyvium(DirectModeFunctions, GenericFunctions, MethodModeFunctions):
    pass



class DeviceNotConnectedToIviumSoftError(Exception):
    '''Raised when your device is not connected in the IviumSoft software'''

    def __init__(self, message="IviumSoft has no connected device. Please, connect one."):
        self.message = message
        super().__init__(self.message)



from .direct_mode_functions import DirectModeFunctions
from .generic_functions import GenericFunctions
from .method_mode_functions import MethodModeFunctions
from .we32_functions import We32Functions


class Core(DirectModeFunctions, GenericFunctions, MethodModeFunctions, We32Functions):
    pass



'''The module provides base class for shared state and methods related to Ivium driver.'''
from typing import Any

from cffi import FFI

from ..util import get_ivium_dll_path

ffi = FFI()


class CoreBase:
    """
    Base class for providing shared state and methods related to Ivium driver.
    """
    __is_driver_open = False
    __lib = ffi.dlopen(get_ivium_dll_path())

    @staticmethod
    def get_lib() -> Any:
        """
        Returns the library instance.
        """
        return CoreBase.__lib

    @staticmethod
    def is_driver_open() -> bool:
        """
        Returns the driver open status.
        """
        return CoreBase.__is_driver_open

    @staticmethod
    def set_driver_open(status: bool) -> None:
        """
        Sets the driver open status.

        :param status: Driver open status as a boolean.
        """
        CoreBase.__is_driver_open = status
